"""Core banded matrix definitions and functions."""

# Copyright 2013 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

import bandmat.full as fl

import numpy as np

cimport numpy as cnp
cimport cython

cnp.import_array()
cnp.import_ufunc()

class BandMat(object):
    """A memory-efficient representation of a square banded matrix.

    An N by N matrix with bandwidth D can be stored efficiently by storing its
    band as a rectangular D by N matrix.
    This class is a lightweight wrapper around a rectangular matrix being used
    in this way, and stores additional details needed to recover the square
    matrix such as the lower and upper bandwidths.

    The representation used for the rectangular matrix is the same one used by
    BLAS and LAPACK (and thus scipy): the columns of the rectangular matrix are
    (parts of) the columns of the square matrix being represented, and the
    successive rows of the rectangular matrix give the superdiagonals, diagonal
    and subdiagonals in order (starting with the outermost superdiagonal and
    ending with the outermost subdiagonal).
    See the "Band Storage" section of the LAPACK Users' Guide at
    http://www.netlib.org/lapack/lug/node124.html or the docstring for
    scipy.linalg.solve_banded for some examples.

    `l` is the number of subdiagonals stored and `u` is the number of
    superdiagonals stored.
    Thus `l` and `u` determine the band in which the entries of the represented
    matrix can be non-zero.
    `data` is the LAPACK-style banded matrix representation of the square
    matrix (or of the transpose of the square matrix if `transposed` is True)
    stored as a numpy array.
    Note that if `transposed` is True, `l` and `u` still refer to the square
    matrix being represented rather than to its transpose.
    """
    def __init__(self, l, u, data, transposed = False):
        self.l = l
        self.u = u
        self.data = data
        self.transposed = transposed

        assert self.l >= 0
        assert self.u >= 0
        assert self.data.ndim == 2
        assert self.data.shape[0] == self.l + self.u + 1

    def __repr__(self):
        return ('BandMat(%r, %r, %r, transposed=%r)' %
                (self.l, self.u, self.data, self.transposed))

    @property
    def size(self):
        """Returns the size of this matrix."""
        return self.data.shape[1]

    @property
    def T(self):
        """Returns the transpose of this matrix.

        This is a cheap operation since it just sets a flag internally.
        The returned BandMat has the same underlying data array as `self`.
        """
        return BandMat(self.u, self.l, self.data,
                       transposed = not self.transposed)

    def full(self):
        """Converts this BandMat to a conventional numpy array.

        The returned numpy array represents the same matrix as `self`.
        """
        if self.transposed:
            return fl.band_c(self.u, self.l, self.data).T
        else:
            return fl.band_c(self.l, self.u, self.data)

    def copy_exact(self):
        """Returns a copy of this BandMat.

        The returned BandMat represents the same matrix as `self`, but has a
        newly-created underlying data array.
        It has the same `transposed` setting as `self`.
        """
        return BandMat(self.l, self.u, self.data.copy(),
                       transposed = self.transposed)

    def copy(self):
        """Returns a copy of this BandMat with transposed set to False.

        The returned BandMat represents the same matrix as `self`, but has a
        newly-created underlying data array, and always has `transposed` set to
        False.
        """
        l = self.l
        u = self.u
        if self.transposed:
            return BandMat(l, u, fl.band_cTe(u, l, self.data))
        else:
            return BandMat(l, u, self.data.copy())

    def equiv(self, l_new = None, u_new = None, transposed_new = None,
              zero_extra = False):
        """Returns an equivalent BandMat stored differently.

        The returned BandMat represents the same matrix as `self`, but has a
        newly-created underlying data array, and has possibly different
        parameters `l`, `u` and `transposed`.
        The new values of these parameters are given by `l_new`, `u_new` and
        `transposed_new`, with the corresponding value from `self` used if any
        of these are None.

        If `zero_extra` is True then the underlying data array of the returned
        BandMat is guaranteed to have extra entries set to zero.
        """
        l = self.l
        u = self.u
        if l_new is None:
            l_new = l
        if u_new is None:
            u_new = u
        if transposed_new is None:
            transposed_new = self.transposed

        assert l_new >= l
        assert u_new >= u

        data_new = np.empty((l_new + u_new + 1, self.size))

        ll, uu = (u, l) if transposed_new else (l, u)
        ll_new, uu_new = (u_new, l_new) if transposed_new else (l_new, u_new)

        data_new[(uu_new - uu_new):(uu_new - uu)] = 0.0
        data_new[(uu_new + ll + 1):(uu_new + ll_new + 1)] = 0.0
        data_new_co = data_new[(uu_new - uu):(uu_new + ll + 1)]
        if self.transposed == transposed_new:
            data_new_co[:] = self.data
            if zero_extra:
                fl.zero_extra_entries(ll, uu, data_new_co)
        else:
            fl.band_cTe(uu, ll, self.data, target_rect = data_new_co)

        return BandMat(l_new, u_new, data_new, transposed = transposed_new)

    @cython.boundscheck(False)
    def plus_equals_band_of(self, mat_bm, double mult = 1.0):
        """Adds a multiple of a band of another matrix to this matrix in-place.

        Any entries of `mat_bm` which lie outside of `self` are ignored.
        Thus to implement conventional matrix addition, `self` must be large
        enough to contain the result of the addition, i.e. `self` must have at
        least as many subdiagonals and superdiagonals as `mat_bm`.

        The statement `target_bm.plus_equals_band_of(mat_bm, mult)` where
        `target_bm` and `mat_bm` are BandMats is the equivalent of:

            target_full += band_ec(l, u, mat_full) * mult

        where `target_full` and `mat_full` are square numpy arrays.
        Here `l` is `target_bm.l` and `u` is `target_bm.u`.
        """
        cdef long frames
        cdef long l_a, u_a, transposed_a
        cdef long l_b, u_b, transposed_b
        cdef cnp.ndarray[cnp.float64_t, ndim=2] a_data
        cdef cnp.ndarray[cnp.float64_t, ndim=2] b_data

        l_a = self.l
        u_a = self.u
        a_data = self.data
        transposed_a = self.transposed
        assert l_a >= 0
        assert u_a >= 0
        assert a_data.shape[0] == l_a + u_a + 1

        l_b = mat_bm.l
        u_b = mat_bm.u
        b_data = mat_bm.data
        transposed_b = mat_bm.transposed
        assert l_b >= 0
        assert u_b >= 0
        assert b_data.shape[0] == l_b + u_b + 1

        frames = a_data.shape[1]
        assert b_data.shape[1] == frames

        cdef long o
        cdef unsigned long row_a, row_b
        cdef long d_a, d_b
        cdef unsigned long frame

        for o in range(-min(u_a, u_b), min(l_a, l_b) + 1):
            row_a = (l_a - o) if transposed_a else (u_a + o)
            row_b = (l_b - o) if transposed_b else (u_b + o)
            d_a = o if transposed_a else 0
            d_b = o if transposed_b else 0
            for frame in range(max(0, -o), max(0, frames + min(0, -o))):
                a_data[row_a, frame + d_a] += b_data[row_b, frame + d_b] * mult

        return

    def __add__(self, other):
        """Sums two banded matrices.

        The expression `a_bm + b_bm` where `a_bm` and `b_bm` are BandMats is
        the equivalent of:

            a_full + b_full

        where `a_full` and `b_full` are square numpy arrays.
        """
        if not isinstance(other, BandMat):
            return NotImplemented

        assert self.size == other.size
        c_bm = self.equiv(l_new = max(self.l, other.l),
                          u_new = max(self.u, other.u))
        c_bm.plus_equals_band_of(other)
        return c_bm

    def __sub__(self, other):
        """Subtracts one banded matrix from another.

        The expression `a_bm - b_bm` where `a_bm` and `b_bm` are BandMats is
        the equivalent of:

            a_full - b_full

        where `a_full` and `b_full` are square numpy arrays.
        """
        if not isinstance(other, BandMat):
            return NotImplemented

        assert self.size == other.size
        c_bm = self.equiv(l_new = max(self.l, other.l),
                          u_new = max(self.u, other.u))
        c_bm.plus_equals_band_of(other, mult = -1.0)
        return c_bm

    def __iadd__(self, other):
        """Adds another matrix to this matrix in-place.

        The statement `a_bm += b_bm` where `a_bm` and `b_bm` are BandMats is
        the equivalent of:

            a_full += b_full

        where `a_full` and `b_full` are square numpy arrays.
        """
        if not isinstance(other, BandMat):
            return NotImplemented

        assert self.size == other.size
        assert self.l >= other.l
        assert self.u >= other.u

        self.plus_equals_band_of(other)
        return self

    def __isub__(self, other):
        """Subtracts another matrix from this matrix in-place.

        The statement `a_bm -= b_bm` where `a_bm` and `b_bm` are BandMats is
        the equivalent of:

            a_full -= b_full

        where `a_full` and `b_full` are square numpy arrays.
        """
        if not isinstance(other, BandMat):
            return NotImplemented

        assert self.size == other.size
        assert self.l >= other.l
        assert self.u >= other.u

        self.plus_equals_band_of(other, mult = -1.0)
        return self

    def __pos__(self):
        """Take the positive of a banded matrix.

        The expression `+a_bm` where `a_bm` is a BandMat is the equivalent of:

            +a_full

        where `a_full` is a square numpy array.
        """
        return BandMat(self.l, self.u, +self.data,
                       transposed = self.transposed)

    def __neg__(self):
        """Take the negative of a banded matrix.

        The expression `-a_bm` where `a_bm` is a BandMat is the equivalent of:

            -a_full

        where `a_full` is a square numpy array.
        """
        return BandMat(self.l, self.u, -self.data,
                       transposed = self.transposed)

    def __mul__(self, other):
        """Multiplies a banded matrix by a scalar.

        The expression `a_bm * mult` where `a_bm` is a BandMat is the
        equivalent of:

            a_full * mult

        where `a_full` is a square numpy array.
        """
        try:
            mult = float(other)
        except:
            return NotImplemented

        return BandMat(self.l, self.u, self.data * mult,
                       transposed = self.transposed)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __floordiv__(self, other):
        """Floor-divides a banded matrix by a scalar.

        The expression `a_bm // mult` where `a_bm` is a BandMat is the
        equivalent of:

            a_full // mult

        where `a_full` is a square numpy array.
        """
        try:
            mult = float(other)
        except:
            return NotImplemented

        return BandMat(self.l, self.u, self.data.__floordiv__(mult),
                       transposed = self.transposed)

    def __div__(self, other):
        """Old-style divides a banded matrix by a scalar.

        When using old-style division (c.f. `from __future__ import division`),
        the expression `a_bm / mult` where `a_bm` is a BandMat is the
        equivalent of:

            a_full / mult

        where `a_full` is a square numpy array.
        """
        try:
            mult = float(other)
        except:
            return NotImplemented

        return BandMat(self.l, self.u, self.data.__div__(mult),
                       transposed = self.transposed)

    def __truediv__(self, other):
        """Divides a banded matrix by a scalar.

        When using new-style division (c.f. `from __future__ import division`),
        the expression `a_bm / mult` where `a_bm` is a BandMat is the
        equivalent of:

            a_full / mult

        where `a_full` is a square numpy array.
        """
        try:
            mult = float(other)
        except:
            return NotImplemented

        return BandMat(self.l, self.u, self.data.__truediv__(mult),
                       transposed = self.transposed)

    def __imul__(self, other):
        """Multiplies this matrix by a scalar in-place.

        The statement `a_bm *= mult` where `a_bm` is a BandMat is the
        equivalent of:

            a_full *= mult

        where `a_full` is a square numpy array.
        """
        try:
            mult = float(other)
        except:
            return NotImplemented

        self.data *= mult
        return self

    def __ifloordiv__(self, other):
        """Floor-divides this matrix by a scalar in-place.

        The statement `a_bm //= mult` where `a_bm` is a BandMat is the
        equivalent of:

            a_full //= mult

        where `a_full` is a square numpy array.
        """
        try:
            mult = float(other)
        except:
            return NotImplemented

        self.data.__ifloordiv__(mult)
        return self

    def __idiv__(self, other):
        """Old-style divides this matrix by a scalar in-place.

        When using old-style division (c.f. `from __future__ import division`),
        the expression `a_bm /= mult` where `a_bm` is a BandMat is the
        equivalent of:

            a_full /= mult

        where `a_full` is a square numpy array.
        """
        try:
            mult = float(other)
        except:
            return NotImplemented

        self.data.__itruediv__(mult)
        return self

    def __itruediv__(self, other):
        """Divides this matrix by a scalar in-place.

        When using new-style division (c.f. `from __future__ import division`),
        the expression `a_bm /= mult` where `a_bm` is a BandMat is the
        equivalent of:

            a_full /= mult

        where `a_full` is a square numpy array.
        """
        try:
            mult = float(other)
        except:
            return NotImplemented

        self.data.__itruediv__(mult)
        return self

    def reverse_view(self):
        """Returns the "time-reversed" version of this matrix.

        This function is implemented by taking a view of the underlying data
        array.
        To obtain a BandMat with a fresh underlying data array, `.copy_exact()`
        should be called on the result.

        The expression `a_bm.reverse_view()` where `a_bm` is a BandMat is the
        equivalent of:

            a_full[::-1, ::-1]

        where `a_full` is a square numpy array.
        """
        return BandMat(self.u, self.l, self.data[::-1, ::-1],
                       transposed = self.transposed)

    def sub_matrix_view(self, start, end):
        """Returns a sub-matrix of this matrix.

        This is implemented by taking a view of the underlying data array.
        To obtain a BandMat with a fresh underlying data array, `.copy_exact()`
        should be called on the result.

        The expression `a_bm.sub_matrix_view[start, end]` where `a_bm` is a
        BandMat is the equivalent of:

            a_full[start:end, start:end]

        where `a_full` is a square numpy array.
        """
        assert 0 <= start <= end <= self.size
        return BandMat(self.l, self.u, self.data[:, start:end],
                       transposed = self.transposed)

    def embed_as_sub_matrix(self, start, size):
        """Returns a new matrix containing this matrix as a sub-matrix.

        The statement `b_bm = a_bm.embed_as_sub_matrix(start, size)` where
        `a_bm` is a BandMat is the equivalent of:

            end = start + len(a_full)
            b_full = np.zeros((size, size))
            b_full[start:end, start:end] = a_full

        where `a_full` is a square numpy array.
        """
        end = start + self.size
        assert 0 <= start <= end <= size

        l, u = self.l, self.u
        ll, uu = (u, l) if self.transposed else (l, u)

        data = np.empty((l + u + 1, size))
        data[:, 0:start] = 0.0
        data[:, start:end] = self.data
        data[:, end:size] = 0.0
        fl.zero_extra_entries(ll, uu, data[:, start:end])
        return BandMat(self.l, self.u, data, transposed = self.transposed)

def zeros(l, u, size):
    """Returns the zero matrix as a BandMat.

    The returned BandMat `ret_bm` has `ret_bm.l = l`, `ret_bm.u` = `u` and
    `ret_bm.size = size`.
    """
    data = np.zeros((l + u + 1, size))
    return BandMat(l, u, data)

def from_full(l, u, mat_full):
    """Converts a square banded numpy array to a BandMat.

    The returned BandMat represents the same matrix as `mat_full`.
    `mat_full` should be a numpy array representing a square matrix with zeros
    outside the band specified by `l` and `u`.
    An AssertionError is raised if `mat_full` has non-zero entries outside the
    specified band.
    """
    mat_bm = BandMat(l, u, fl.band_e(l, u, mat_full))
    # check `mat_full` is zero outside the specified band
    assert np.all(mat_bm.full() == mat_full)
    return mat_bm

def band_c_bm(l, u, mat_rect):
    """Constructs a BandMat from its band.

    The expression `band_c_bm(l, u, mat_rect)` where `mat_rect` is a
    rectangular numpy array is the equivalent of:

        band_c(l, u, mat_rect)

    where the returned value is a square numpy array.
    """
    return BandMat(l, u, mat_rect.copy())

def band_e_bm(l, u, mat_bm):
    """Extracts a band of a BandMat.

    The band to extract is specified by `l` and `u`.

    The expression `band_e_bm(l, u, mat_bm)` where `mat_bm` is a BandMat is the
    equivalent of:

        band_e(l, u, mat_full)

    where `mat_full` is a square numpy array.
    """
    mat_bm_co = band_ec_bm_view(l, u, mat_bm)
    mat_bm_new = mat_bm_co.equiv(l_new = l, u_new = u,
                                 transposed_new = False,
                                 zero_extra = True)
    return mat_bm_new.data

band_ce_bm = fl.band_ce

def band_ec_bm_view(l, u, mat_bm):
    """Effectively applies `band_e_bm` then `band_c_bm`, sharing data arrays.

    The combined operation has the effect of zeroing the entries outside the
    band specified by `l` and `u`.
    This is implemented by taking a view of `mat_bm`'s underlying data array.
    To obtain a BandMat with a fresh underlying data array, `.copy_exact()`
    should be called on the result.
    """
    assert l >= 0
    assert u >= 0

    l_in = mat_bm.l
    u_in = mat_bm.u
    l_out = min(l, l_in)
    u_out = min(u, u_in)
    if mat_bm.transposed:
        return BandMat(
            l_out, u_out,
            mat_bm.data[(l_in - l_out):(l_in + u_out + 1)],
            transposed = True
        )
    else:
        return BandMat(
            l_out, u_out,
            mat_bm.data[(u_in - u_out):(u_in + l_out + 1)]
        )

def band_ec_bm(l, u, mat_bm):
    """Effectively applies `band_e_bm` then `band_c_bm`.

    The combined operation has the effect of zeroing the entries outside the
    band specified by `l` and `u`.

    The expression `band_ec_bm(l, u, mat_bm)` where `mat_bm` is a BandMat is
    the equivalent of:

        band_ec(l, u, mat_full)

    where `mat_full` and the returned value are square numpy arrays.
    """
    return band_ec_bm_view(l, u, mat_bm).copy_exact()

def band_e_bm_common(*mat_bms):
    """Extracts the band from several BandMats in a common way.

    The same band is used for each of the BandMats, and this band is the
    smallest band that encompasses the bands of all the BandMats.
    This allows BandMats to be compared with the same tools used to compare
    conventional numpy arrays (and without having to explicitly construct the
    corresponding full matrices).

    Example usage:
    >>> import bandmat as bm
    >>> cc = bm.band_e_bm_common
    >>> def allclose(x, y): return x.shape == y.shape and np.allclose(x, y)
    >>> a_bm = bm.zeros(3, 2, 10)
    >>> b_bm = bm.zeros(1, 3, 10).T
    >>> # these represent the same matrix
    >>> allclose(a_bm.full(), b_bm.full())
    True
    >>> # naive comparison doesn't show this
    >>> allclose(a_bm.data, b_bm.data)
    False
    >>> # comparison after normalizing works
    >>> allclose(*cc(a_bm, b_bm))
    True
    """
    if not mat_bms:
        return []
    else:
        l = max([ mat_bm.l for mat_bm in mat_bms ])
        u = max([ mat_bm.u for mat_bm in mat_bms ])
        return [ band_e_bm(l, u, mat_bm) for mat_bm in mat_bms ]

def diag(vec_or_mat_bm):
    """Constructs a diagonal BandMat or extracts the diagonal from a BandMat.

    If `vec_or_mat_bm` is a vector `vec` (i.e. a one-dimensional numpy array),
    then this function returns a BandMat representing the diagonal matrix with
    diagonal `vec`.
    If `vec_or_mat_bm` is a BandMat `mat_bm`, then this function returns the
    vector consisting of the diagonal of the matrix represented by `mat_bm`.
    In both cases the array storing the vector and the BandMat's underlying
    data array share memory.

    The expression `diag(vec_or_mat_bm)` where `vec_or_mat_bm` is a vector or a
    BandMat is the equivalent of:

        np.diag(vec_or_mat_full)

    where `vec_or_mat_full` is a vector or a square numpy array.
    """
    if not isinstance(vec_or_mat_bm, BandMat):
        vec = vec_or_mat_bm
        assert vec.ndim == 1
        return BandMat(0, 0, np.reshape(vec, (1, -1)))
    else:
        mat_bm = vec_or_mat_bm
        row = mat_bm.l if mat_bm.transposed else mat_bm.u
        return mat_bm.data[row]



"""Addition, multiplication, etc using banded matrices."""

# Copyright 2013 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

import bandmat.core as bm_core

import numpy as np

cimport numpy as cnp
cimport cython

cnp.import_array()
cnp.import_ufunc()

@cython.boundscheck(False)
def dot_mv_plus_equals(a_bm,
                       cnp.ndarray[cnp.float64_t, ndim=1] b,
                       cnp.ndarray[cnp.float64_t, ndim=1] target):
    """Multiplies a banded matrix by a vector, adding the result to a vector.

    The statement `dot_mv_plus_equals(a_bm, b, target)` where `a_bm` is a
    BandMat is the equivalent of:

        target += np.dot(a_full, b)

    where `a_full` is a square numpy array.
    """
    # (FIXME : could wrap corresponding BLAS routine (gbmv) instead)
    cdef long frames
    cdef long l_a, u_a, transposed
    cdef cnp.ndarray[cnp.float64_t, ndim=2] a_data

    l_a = a_bm.l
    u_a = a_bm.u
    a_data = a_bm.data
    transposed = a_bm.transposed
    assert l_a >= 0
    assert u_a >= 0
    assert a_data.shape[0] == l_a + u_a + 1

    frames = a_data.shape[1]
    assert b.shape[0] == frames
    assert target.shape[0] == frames

    cdef long o_a
    cdef unsigned long row_a
    cdef long d_a
    cdef unsigned long frame

    for o_a in range(-u_a, l_a + 1):
        row_a = (l_a - o_a) if transposed else (u_a + o_a)
        d_a = 0 if transposed else -o_a
        for frame in range(max(0, o_a), max(0, frames + min(0, o_a))):
            target[frame] += (
                a_data[row_a, frame + d_a] *
                b[frame - o_a]
            )

    return

def dot_mv(a_bm, b):
    """Multiplies a banded matrix by a vector.

    The expression `dot_mv(a_bm, b)` where `a_bm` is a BandMat is the
    equivalent of:

        np.dot(a_full, b)

    where `a_full` is a square numpy array.
    """
    size = len(b)
    assert a_bm.size == size
    c = np.zeros((size,))
    dot_mv_plus_equals(a_bm, b, c)
    return c

@cython.boundscheck(False)
def dot_mm_plus_equals(a_bm, b_bm, target_bm,
                       cnp.ndarray[cnp.float64_t, ndim=1] diag = None):
    """Multiplies two banded matrices, adding the result to a banded matrix.

    If `diag` is None, computes A . B, where . indicates matrix multiplication,
    and adds the result to `target_bm`.
    If `diag` is not None, computes A . D . B, where D is the diagonal matrix
    with diagonal `diag`, and adds result to `target_bm`.

    If `target_bm` does not contain enough rows to contain the result of A . B,
    then only diagonals of A . B which contribute to the rows present in the
    banded representation used by `target_bm` are computed.
    This is more efficient in the case that not all diagonals of `target_bm`
    are needed.

    The statement `dot_mm_plus_equals(a_bm, b_bm, target_bm, diag=None)` where
    `a_bm`, `b_bm` and `target_bm` are BandMats is the equivalent of:

        target_full += band_ec(l, u, np.dot(a_full, b_full))

    where `a_full`, `b_full` and `target_full` are square numpy arrays.
    Here `l` is `target_bm.l` and `u` is `target_bm.u`.
    If `diag` is not None then it is the equivalent of:

        target_full += (
            band_ec(l, u, np.dot(np.dot(a_full, np.diag(diag)), b_full))
        )
    """
    cdef long frames
    cdef long l_a, u_a, transposed_a
    cdef long l_b, u_b, transposed_b
    cdef long l_c, u_c, transposed_c
    cdef long use_diag
    cdef cnp.ndarray[cnp.float64_t, ndim=2] a_data
    cdef cnp.ndarray[cnp.float64_t, ndim=2] b_data
    cdef cnp.ndarray[cnp.float64_t, ndim=2] c_data

    l_a = a_bm.l
    u_a = a_bm.u
    a_data = a_bm.data
    transposed_a = a_bm.transposed
    assert l_a >= 0
    assert u_a >= 0
    assert a_data.shape[0] == l_a + u_a + 1

    l_b = b_bm.l
    u_b = b_bm.u
    b_data = b_bm.data
    transposed_b = b_bm.transposed
    assert l_b >= 0
    assert u_b >= 0
    assert b_data.shape[0] == l_b + u_b + 1

    l_c = target_bm.l
    u_c = target_bm.u
    c_data = target_bm.data
    transposed_c = target_bm.transposed
    assert l_c >= 0
    assert u_c >= 0
    assert c_data.shape[0] == l_c + u_c + 1

    frames = a_data.shape[1]
    assert b_data.shape[1] == frames
    assert c_data.shape[1] == frames

    use_diag = (diag is not None)
    if use_diag:
        assert diag.shape[0] == frames

    cdef long o_a, o_b, o_c
    cdef unsigned long row_a, row_b, row_c
    cdef long d_a, d_b, d_c
    cdef unsigned long frame

    for o_c in range(-min(u_c, u_a + u_b), min(l_c, l_a + l_b) + 1):
        for o_a in range(-min(u_a, l_b - o_c), min(l_a, u_b + o_c) + 1):
            o_b = o_c - o_a
            row_a = (l_a - o_a) if transposed_a else (u_a + o_a)
            row_b = (l_b - o_b) if transposed_b else (u_b + o_b)
            row_c = (l_c - o_c) if transposed_c else (u_c + o_c)
            d_a = o_a if transposed_a else 0
            d_b = 0 if transposed_b else -o_b
            d_c = o_a if transposed_c else -o_b
            for frame in range(max(0, -o_a, o_b),
                               max(0, frames + min(0, -o_a, o_b))):
                c_data[row_c, frame + d_c] += (
                    a_data[row_a, frame + d_a] *
                    b_data[row_b, frame + d_b] *
                    (diag[frame] if use_diag else 1.0)
                )

    return

def dot_mm(a_bm, b_bm, diag = None):
    """Multiplies two banded matrices.

    If `diag` is None, computes A . B, where . indicates matrix multiplication.
    If `diag` is not None, computes A . D . B, where D is the diagonal matrix
    with diagonal `diag`.

    The expression `dot_mm(a_bm, b_bm, diag=None)` where `a_bm` and `b_bm` are
    BandMats is the equivalent of:

        np.dot(a_full, b_full)

    where `a_full` and `b_full` are square numpy arrays.
    If `diag` is not None then it is the equivalent of:

        np.dot(np.dot(a_full, np.diag(diag)), b_full)
    """
    assert a_bm.size == b_bm.size
    c_bm = bm_core.zeros(a_bm.l + b_bm.l, a_bm.u + b_bm.u, a_bm.size)
    dot_mm_plus_equals(a_bm, b_bm, c_bm, diag = diag)
    return c_bm

def dot_mm_partial(l, u, a_bm, b_bm, diag = None):
    """Computes part of the result of multiplying two banded matrices.

    If `diag` is None, computes part of C = A . B, where . indicates matrix
    multiplication.
    If `diag` is not None, computes part of C = A . D . B, where D is the
    diagonal matrix with diagonal `diag`.

    This function only computes the diagonals of C that are within the band
    specified by `l` and `u`.
    This is desirable in the case that not all diagonals of C are needed.

    The expression `dot_mm_partial(l, u, a_bm, b_bm, diag=None)` where `a_bm`
    and `b_bm` are BandMats is the equivalent of:

        band_ec(l, u, np.dot(a_full, b_full))

    where `a_full` and `b_full` are square numpy arrays.
    If `diag` is not None then it is the equivalent of:

        band_ec(l, u, np.dot(np.dot(a_full, np.diag(diag)), b_full))
    """
    assert a_bm.size == b_bm.size
    c_bm = bm_core.zeros(l, u, a_bm.size)
    dot_mm_plus_equals(a_bm, b_bm, c_bm, diag = diag)
    return c_bm

def dot_mmm_partial(l, u, a_bm, b_bm, c_bm):
    """Computes part of the result of multiplying three banded matrices.

    Computes D = A . (B . C), where . indicates matrix multiplication.

    This function only computes the diagonals of D that are within the band
    specified by `l` and `u`.
    Furthermore the intermediate result E = B . C is also only partially
    computed, with rows that can no have influence on the final result never
    being computed.
    This behavior is desirable in the case that not all diagonals of D are
    needed.

    The expression `dot_mmm_partial(l, u, a_bm, b_bm, c_bm)` where `a_bm`,
    `b_bm` and `c_bm` are BandMats is the equivalent of:

        band_ec(l, u, np.dot(a_full, np.dot(b_full, c_full)))

    where `a_full`, `b_full` and `c_full` are square numpy arrays.
    """
    l_i = l + a_bm.u
    u_i = u + a_bm.l
    return dot_mm_partial(l, u, a_bm, dot_mm_partial(l_i, u_i, b_bm, c_bm))

@cython.boundscheck(False)
def band_of_outer_plus_equals(cnp.ndarray[cnp.float64_t, ndim=1] a_vec,
                              cnp.ndarray[cnp.float64_t, ndim=1] b_vec,
                              target_bm,
                              double mult = 1.0):
    """Adds the band of the outer product of two vectors to a banded matrix.

    The statement `band_of_outer_plus_equals(a_vec, b_vec, target_bm, mult)`
    where `target_bm` is a BandMat is the equivalent of:

        target_full += band_ec(l, u, np.outer(a_vec, b_vec) * mult)

    where `target_full` is a square numpy array.
    Here `l` is `target_bm.l` and `u` is `target_bm.u`.
    """
    cdef long frames
    cdef long l_c, u_c, transposed_c
    cdef cnp.ndarray[cnp.float64_t, ndim=2] c_data

    l_c = target_bm.l
    u_c = target_bm.u
    c_data = target_bm.data
    transposed_c = target_bm.transposed
    assert l_c >= 0
    assert u_c >= 0
    assert c_data.shape[0] == l_c + u_c + 1

    frames = c_data.shape[1]
    assert a_vec.shape[0] == frames
    assert b_vec.shape[0] == frames

    cdef long o_c
    cdef unsigned long row_c
    cdef long d_c
    cdef unsigned long frame

    for o_c in range(-u_c, l_c + 1):
        row_c = (l_c - o_c) if transposed_c else (u_c + o_c)
        d_c = o_c if transposed_c else 0
        for frame in range(max(0, -o_c), max(0, frames + min(0, -o_c))):
            c_data[row_c, frame + d_c] += (
                a_vec[frame + o_c] * b_vec[frame] * mult
            )

    return

def trace_dot(a_bm, b_bm):
    """Convenience method to compute the matrix inner product.

    The expression `trace_dot(a_bm, b_bm)` where `a_bm` and `b_bm` are BandMats
    is the equivalent of:

        np.trace(np.dot(a_full.T, b_full))

    where `a_full` and `b_full` are square numpy arrays.

    This operation is a valid inner product over the vector space of square
    matrices of a given size.
    """
    return np.sum(bm_core.diag(dot_mm_partial(0, 0, a_bm.T, b_bm)))



"""Linear algebra operations for banded matrices.

Several of the functions in this module deal with symmetric banded matrices.
Symmetric banded matrices can in principle be stored even more efficiently than
general banded matrices, by not storing the superdiagonals (or subdiagonals).
However in bandmat the simpler, more explicit representation which stores the
whole band is used instead.
This is slightly less efficient in terms of memory and time, but results in a
more unified interface and simpler code in the author's limited experience.
Note that this is in line with the convention used by numpy and scipy for
symmetric "full" (i.e. non-banded) matrices, where both the upper and lower
triangles are represented in memory.
"""

# Copyright 2013 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

import bandmat as bm
from bandmat import BandMat

import numpy as np
import scipy.linalg as sla

cimport numpy as cnp
cimport cython

cnp.import_array()
cnp.import_ufunc()

def cholesky(mat_bm, lower = False, alternative = False):
    """Computes the Cholesky factor of a positive definite banded matrix.

    The conventional Cholesky decomposition of a positive definite matrix P is
    P = L . L.T, where the lower Cholesky factor L is lower-triangular, the
    upper Cholesky factor L.T is upper-triangular, and . indicates matrix
    multiplication.
    Each positive definite matrix has a unique Cholesky decomposition.
    Given a positive definite banded matrix, this function computes its
    Cholesky factor, which is also a banded matrix.

    This function can also work with the alternative Cholesky decomposition.
    The alternative Cholesky decomposition of P is P = L.T . L, where again L
    is lower-triangular.
    Whereas the conventional Cholesky decomposition is intimately connected
    with linear Gaussian autoregressive probabilistic models defined forwards
    in time, the alternative Cholesky decomposition is intimately connected
    with linear Gaussian autoregressive probabilistic models defined backwards
    in time.

    `mat_bm` (representing P above) should represent the whole matrix, not just
    the lower or upper triangles.
    If the matrix represented by `mat_bm` is not symmetric then the behavior of
    this function is undefined (currently either the upper or lower triangle is
    used to compute the Cholesky factor and the rest of the matrix is ignored).
    If `lower` is True then the lower Cholesky factor is returned, and
    otherwise the upper Cholesky factor is returned.
    If `alternative` is True then the Cholesky factor returned is for the
    alternative Cholesky decomposition, and otherwise it is for the
    conventional Cholesky decomposition.
    """
    if mat_bm.transposed:
        mat_bm = mat_bm.T

    depth = mat_bm.l
    assert mat_bm.u == depth
    assert depth >= 0

    l = depth if lower else 0
    u = 0 if lower else depth

    if mat_bm.size == 0:
        chol_bm = BandMat(l, u, np.zeros((depth + 1, 0)))
    else:
        mat_half_data = mat_bm.data[(depth - u):(depth + l + 1)]
        if alternative:
            chol_data = sla.cholesky_banded(mat_half_data[::-1, ::-1],
                                            lower = not lower)[::-1, ::-1]
        else:
            chol_data = sla.cholesky_banded(mat_half_data, lower = lower)
        chol_bm = BandMat(l, u, chol_data)

    return chol_bm

def cho_solve(chol_bm, b):
    """Solves a matrix equation given the Cholesky decomposition of the matrix.

    Solves A . x = b for x, where A is a positive definite banded matrix, x and
    b are vectors, and . indicates matrix multiplication.
    `chol_bm` is a Cholesky factor of A (either upper or lower).
    """
    assert chol_bm.size == len(b)

    if chol_bm.transposed:
        chol_bm = chol_bm.T

    lower = (chol_bm.u == 0)
    x = sla.cho_solve_banded((chol_bm.data, lower), b)
    return x

def solve(a_bm, b):
    """Solves a matrix equation.

    Solves A . x = b for x, where A is a square banded matrix, x and b are
    vectors, and . indicates matrix multiplication.
    """
    assert a_bm.size == len(b)

    # below is necessary since sla.solve_banded does not have a transpose flag,
    #   and so is not capable of working with the transposed matrix directly.
    #   (In fact (surprisingly!) the underlying LAPACK function dgbsv does not
    #   have a transpose flag either, though gbsv calls dgbtrf and dgbtrs, and
    #   dgbtrs does have a transpose flag, so LAPACK is capable of working with
    #   the transposed matrix directly in principle).
    if a_bm.transposed:
        a_bm = a_bm.equiv(transposed_new = False)

    if a_bm.size == 0:
        x = np.zeros_like(b)
    else:
        x = sla.solve_banded((a_bm.l, a_bm.u), a_bm.data, b)
    return x

def solveh(a_bm, b):
    """Solves a positive definite matrix equation.

    Solves A . x = b for x, where A is a positive definite banded matrix, x and
    b are vectors, and . indicates matrix multiplication.

    `a_bm` (representing A above) should represent the whole matrix, not just
    the lower or upper triangles.
    If the matrix represented by `a_bm` is not symmetric then the behavior of
    this function is undefined (currently either the upper or lower triangle is
    used and the rest of the matrix is ignored).
    """
    assert a_bm.size == len(b)

    if a_bm.transposed:
        a_bm = a_bm.T

    depth = a_bm.l
    assert a_bm.u == depth
    assert depth >= 0

    lower = False
    l = depth if lower else 0
    u = 0 if lower else depth

    if a_bm.size == 0:
        x = np.zeros_like(b)
    else:
        a_half_data = a_bm.data[(depth - u):(depth + l + 1)]
        x = sla.solveh_banded(a_half_data, b, lower = lower)
    return x

@cython.boundscheck(False)
def band_of_inverse_from_chol(chol_bm):
    """Computes the band of the inverse of a positive definite banded matrix.

    Computes the band of the inverse of a positive definite banded matrix given
    its Cholesky factor.
    Equivalently, finds the band of the covariance matrix of a discrete time
    process which is linear-Gaussian backwards in time.

    `chol_bm` can be either the lower or upper Cholesky factor.
    """
    assert chol_bm.l == 0 or chol_bm.u == 0

    if chol_bm.u != 0:
        chol_bm = chol_bm.T

    cdef unsigned long depth
    cdef long frames
    cdef long l_chol, u_chol, transposed_chol
    cdef cnp.ndarray[cnp.float64_t, ndim=2] chol_data
    cdef cnp.ndarray[cnp.float64_t, ndim=2] cov_data

    l_chol = chol_bm.l
    u_chol = chol_bm.u
    chol_data = chol_bm.data
    transposed_chol = chol_bm.transposed
    assert l_chol >= 0
    assert u_chol == 0
    assert chol_data.shape[0] == l_chol + u_chol + 1

    depth = l_chol
    frames = chol_data.shape[1]
    cov_data = np.zeros((depth * 2 + 1, frames))

    cdef long frame_l
    cdef unsigned long frame
    cdef long curr_depth
    cdef long k_1, k_2
    cdef unsigned long row_chol
    cdef long d_chol
    cdef double mult

    curr_depth = 0
    for frame_l in range(frames - 1, -1, -1):
        frame = frame_l
        row_chol = l_chol if transposed_chol else u_chol
        mult = 1.0 / chol_data[row_chol, frame]
        cov_data[depth, frame] = mult * mult
        for k_2 in range(curr_depth, -1, -1):
            for k_1 in range(curr_depth, 0, -1):
                row_chol = l_chol - k_1 if transposed_chol else u_chol + k_1
                d_chol = k_1 if transposed_chol else 0
                cov_data[depth + k_2, frame] -= (
                    chol_data[row_chol, frame + d_chol] *
                    cov_data[depth + k_1 - k_2, frame + k_2] *
                    mult
                )
        for k_2 in range(curr_depth, 0, -1):
            cov_data[depth - k_2, frame + k_2] = (
                cov_data[depth + k_2, frame]
            )
        if curr_depth < depth:
            curr_depth += 1

    cov_bm = BandMat(depth, depth, cov_data)
    return cov_bm

def band_of_inverse(mat_bm):
    """Computes band of the inverse of a positive definite banded matrix."""
    depth = mat_bm.l
    assert mat_bm.u == depth
    chol_bm = cholesky(mat_bm, lower = True)
    band_of_inv_bm = band_of_inverse_from_chol(chol_bm)
    return band_of_inv_bm



"""Tests for addition, multiplication, etc using banded matrices."""

# Copyright 2013 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

from bandmat.testhelp import assert_allclose, get_array_mem
from bandmat.test_core import gen_BandMat

import bandmat as bm
import bandmat.full as fl

import unittest
import numpy as np
import random
from numpy.random import randn, randint

def randBool():
    return randint(0, 2) == 0

class TestTensor(unittest.TestCase):
    def test_dot_mv_plus_equals(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b = randn(size)
            c = randn(size)
            a_full = a_bm.full()
            c_good = c.copy()
            array_mem = get_array_mem(a_bm.data, b, c)

            bm.dot_mv_plus_equals(a_bm, b, c)
            c_good += np.dot(a_full, b)
            assert_allclose(c, c_good)
            assert get_array_mem(a_bm.data, b, c) == array_mem

    def test_dot_mv(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b = randn(size)
            a_full = a_bm.full()

            c = bm.dot_mv(a_bm, b)
            c_good = np.dot(a_full, b)
            assert_allclose(c, c_good)
            assert not np.may_share_memory(c, a_bm.data)
            assert not np.may_share_memory(c, b)

    def test_dot_mm_plus_equals(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            c_bm = gen_BandMat(size)
            diag = None if randBool() else randn(size)
            diag_value = np.ones((size,)) if diag is None else diag
            a_full = a_bm.full()
            b_full = b_bm.full()
            c_full = c_bm.full()
            l = c_bm.l
            u = c_bm.u
            array_mem = get_array_mem(a_bm.data, b_bm.data, c_bm.data)
            if diag is not None:
                diag_mem = get_array_mem(diag)

            bm.dot_mm_plus_equals(a_bm, b_bm, c_bm, diag = diag)
            c_full += fl.band_ec(
                l, u,
                np.dot(np.dot(a_full, np.diag(diag_value)), b_full)
            )
            assert_allclose(c_bm.full(), c_full)
            assert get_array_mem(a_bm.data, b_bm.data, c_bm.data) == array_mem
            if diag is not None:
                assert get_array_mem(diag) == diag_mem

    def test_dot_mm(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            diag = None if randBool() else randn(size)
            diag_value = np.ones((size,)) if diag is None else diag
            a_full = a_bm.full()
            b_full = b_bm.full()

            c_bm = bm.dot_mm(a_bm, b_bm, diag = diag)
            c_full = np.dot(np.dot(a_full, np.diag(diag_value)), b_full)
            assert c_bm.l == a_bm.l + b_bm.l
            assert c_bm.u == a_bm.u + b_bm.u
            assert c_bm.size == size
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)
            assert not np.may_share_memory(c_bm.data, b_bm.data)
            if diag is not None:
                assert not np.may_share_memory(c_bm.data, diag)

    def test_dot_mm_partial(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            diag = None if randBool() else randn(size)
            diag_value = np.ones((size,)) if diag is None else diag
            a_full = a_bm.full()
            b_full = b_bm.full()

            c_bm = bm.dot_mm_partial(l, u, a_bm, b_bm, diag = diag)
            c_full = fl.band_ec(
                l, u,
                np.dot(np.dot(a_full, np.diag(diag_value)), b_full)
            )
            assert c_bm.l == l
            assert c_bm.u == u
            assert c_bm.size == size
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)
            assert not np.may_share_memory(c_bm.data, b_bm.data)
            if diag is not None:
                assert not np.may_share_memory(c_bm.data, diag)

    def test_dot_mmm_partial(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            c_bm = gen_BandMat(size)
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            a_full = a_bm.full()
            b_full = b_bm.full()
            c_full = c_bm.full()

            d_bm = bm.dot_mmm_partial(l, u, a_bm, b_bm, c_bm)
            d_full = fl.band_ec(l, u, np.dot(a_full, np.dot(b_full, c_full)))
            assert d_bm.l == l
            assert d_bm.u == u
            assert d_bm.size == size
            assert_allclose(d_bm.full(), d_full)
            assert not np.may_share_memory(d_bm.data, a_bm.data)
            assert not np.may_share_memory(d_bm.data, b_bm.data)
            assert not np.may_share_memory(d_bm.data, c_bm.data)

    def test_band_of_outer_plus_equals(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_vec = randn(size)
            b_vec = randn(size)
            mult = randn()
            mat_bm = gen_BandMat(size)
            mat_full = mat_bm.full()
            l = mat_bm.l
            u = mat_bm.u
            array_mem = get_array_mem(a_vec, b_vec, mat_bm.data)

            bm.band_of_outer_plus_equals(a_vec, b_vec, mat_bm, mult = mult)
            mat_full += fl.band_ec(l, u, np.outer(a_vec, b_vec) * mult)
            assert_allclose(mat_bm.full(), mat_full)
            assert get_array_mem(a_vec, b_vec, mat_bm.data) == array_mem

    def test_trace_dot(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            a_full = a_bm.full()
            b_full = b_bm.full()

            c = bm.trace_dot(a_bm, b_bm)
            c_good = np.trace(np.dot(a_full.T, b_full))
            assert_allclose(c, c_good)

if __name__ == '__main__':
    unittest.main()



"""Tests for core banded matrix definitions and functions."""

# Copyright 2013 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

from bandmat.testhelp import assert_allclose, assert_allequal, get_array_mem

import bandmat as bm
import bandmat.full as fl

import unittest
import doctest
import numpy as np
import random
from numpy.random import randn, randint

def randBool():
    return randint(0, 2) == 0

def gen_BandMat(size, l = None, u = None, transposed = None):
    """Generates a random BandMat."""
    if l is None:
        l = random.choice([0, 1, randint(0, 10)])
    if u is None:
        u = random.choice([0, 1, randint(0, 10)])
    data = randn(l + u + 1, size)
    if transposed is None:
        transposed = randBool()
    return bm.BandMat(l, u, data, transposed = transposed)

def load_tests(loader, tests, ignore):
    # package-level doctests (N.B. includes other modules, not just core)
    tests.addTests(doctest.DocTestSuite(bm))
    tests.addTests(doctest.DocTestSuite(bm.core))
    return tests

class TestCore(unittest.TestCase):
    def test_BandMat_basic(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)

            assert a_bm.size == size

    def test_BandMat_full(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            l, u = a_bm.l, a_bm.u

            # N.B. these tests are not really testing much of anything (they
            #   are virtually identical to the implementation of BandMat.full),
            #   but this is not that surprising since the lines below are kind
            #   of the definition of the representation used by BandMat in the
            #   two cases (transposed True and transposed False).
            if a_bm.transposed:
                assert_allequal(a_full.T, fl.band_c(u, l, a_bm.data))
            else:
                assert_allequal(a_full, fl.band_c(l, u, a_bm.data))

            assert not np.may_share_memory(a_full, a_bm.data)

    def test_BandMat_T(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)

            assert_allequal(a_bm.T.full(), a_bm.full().T)

            assert a_bm.T.data is a_bm.data
            if size > 0:
                assert np.may_share_memory(a_bm.T.data, a_bm.data)

    def test_BandMat_copy_exact(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_bm = gen_BandMat(size)
            mat_full_orig = mat_bm.full().copy()

            mat_bm_new = mat_bm.copy_exact()
            assert mat_bm_new.l == mat_bm.l
            assert mat_bm_new.u == mat_bm.u
            assert mat_bm_new.transposed == mat_bm.transposed

            # check that copy represents the same matrix
            assert_allequal(mat_bm_new.full(), mat_full_orig)

            # check that copy does not share memory with original
            assert not np.may_share_memory(mat_bm_new.data, mat_bm.data)

            # check that mutating the copy does not change the original
            mat_bm_new.data += 1.0
            assert_allequal(mat_bm.full(), mat_full_orig)

    def test_BandMat_copy(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_bm = gen_BandMat(size)
            mat_full_orig = mat_bm.full().copy()

            mat_bm_new = mat_bm.copy()
            assert mat_bm_new.l == mat_bm.l
            assert mat_bm_new.u == mat_bm.u
            assert not mat_bm_new.transposed

            # check that copy represents the same matrix
            assert_allequal(mat_bm_new.full(), mat_full_orig)

            # check that copy does not share memory with original
            assert not np.may_share_memory(mat_bm_new.data, mat_bm.data)

            # check that mutating the copy does not change the original
            mat_bm_new.data += 1.0
            assert_allequal(mat_bm.full(), mat_full_orig)

    def test_BandMat_equiv(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_bm = gen_BandMat(size)
            l_new = random.choice([None, 0, 1, randint(0, 10)])
            u_new = random.choice([None, 0, 1, randint(0, 10)])
            transposed_new = random.choice([None, True, False])
            zero_extra = randBool()

            l_new_value = mat_bm.l if l_new is None else l_new
            u_new_value = mat_bm.u if u_new is None else u_new
            transposed_new_value = (mat_bm.transposed if transposed_new is None
                                    else transposed_new)

            if l_new_value < mat_bm.l or u_new_value < mat_bm.u:
                self.assertRaises(AssertionError,
                                  mat_bm.equiv,
                                  l_new = l_new, u_new = u_new,
                                  transposed_new = transposed_new,
                                  zero_extra = zero_extra)
            else:
                mat_bm_new = mat_bm.equiv(l_new = l_new, u_new = u_new,
                                          transposed_new = transposed_new,
                                          zero_extra = zero_extra)
                assert mat_bm_new.l == l_new_value
                assert mat_bm_new.u == u_new_value
                assert mat_bm_new.transposed == transposed_new_value
                assert_allequal(mat_bm_new.full(), mat_bm.full())
                assert not np.may_share_memory(mat_bm_new.data, mat_bm.data)

                if zero_extra:
                    mat_new_data_good = (
                        fl.band_e(u_new_value, l_new_value, mat_bm.full().T)
                    ) if mat_bm_new.transposed else (
                        fl.band_e(l_new_value, u_new_value, mat_bm.full())
                    )
                    assert_allequal(mat_bm_new.data, mat_new_data_good)

    def test_BandMat_plus_equals_band_of(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mult = randn()
            target_bm = gen_BandMat(size)
            mat_bm = gen_BandMat(size)
            target_full = target_bm.full()
            mat_full = mat_bm.full()
            array_mem = get_array_mem(target_bm.data, mat_bm.data)

            target_bm.plus_equals_band_of(mat_bm, mult)
            target_full += (
                fl.band_ec(target_bm.l, target_bm.u, mat_full) * mult
            )
            assert_allclose(target_bm.full(), target_full)
            assert get_array_mem(target_bm.data, mat_bm.data) == array_mem

    def test_BandMat_add(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            a_full = a_bm.full()
            b_full = b_bm.full()

            c_bm = a_bm + b_bm
            c_full = a_full + b_full
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)
            assert not np.may_share_memory(c_bm.data, b_bm.data)

            with self.assertRaises(TypeError):
                a_bm + 1.0
            with self.assertRaises(TypeError):
                1.0 + a_bm

    def test_BandMat_sub(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            a_full = a_bm.full()
            b_full = b_bm.full()

            c_bm = a_bm - b_bm
            c_full = a_full - b_full
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)
            assert not np.may_share_memory(c_bm.data, b_bm.data)

            with self.assertRaises(TypeError):
                a_bm - 1.0
            with self.assertRaises(TypeError):
                1.0 - a_bm

    def test_BandMat_iadd(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            a_full = a_bm.full()
            b_full = b_bm.full()
            array_mem = get_array_mem(a_bm.data, b_bm.data)

            if a_bm.l >= b_bm.l and a_bm.u >= b_bm.u:
                a_bm += b_bm
                a_full += b_full
                assert_allclose(a_bm.full(), a_full)
                assert get_array_mem(a_bm.data, b_bm.data) == array_mem
            else:
                with self.assertRaises(AssertionError):
                    a_bm += b_bm

            with self.assertRaises(TypeError):
                a_bm += 1.0

    def test_BandMat_isub(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            a_full = a_bm.full()
            b_full = b_bm.full()
            array_mem = get_array_mem(a_bm.data, b_bm.data)

            if a_bm.l >= b_bm.l and a_bm.u >= b_bm.u:
                a_bm -= b_bm
                a_full -= b_full
                assert_allclose(a_bm.full(), a_full)
                assert get_array_mem(a_bm.data, b_bm.data) == array_mem
            else:
                with self.assertRaises(AssertionError):
                    a_bm -= b_bm

            with self.assertRaises(TypeError):
                a_bm -= 1.0

    def test_BandMat_pos(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()

            b_bm = +a_bm
            b_full = +a_full
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

    def test_BandMat_neg(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()

            b_bm = -a_bm
            b_full = -a_full
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

    def test_BandMat_mul_and_rmul(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mult = randn()
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()

            b_bm = a_bm * mult
            b_full = a_full * mult
            assert b_bm.l == a_bm.l
            assert b_bm.u == a_bm.u
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

            c_bm = mult * a_bm
            c_full = mult * a_full
            assert c_bm.l == a_bm.l
            assert c_bm.u == a_bm.u
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)

            with self.assertRaises(TypeError):
                a_bm * a_bm

    def test_BandMat_various_divs(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mult = randn()
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()

            b_bm = a_bm // mult
            b_full = a_full // mult
            assert b_bm.l == a_bm.l
            assert b_bm.u == a_bm.u
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

            b_bm = a_bm / mult
            b_full = a_full / mult
            assert b_bm.l == a_bm.l
            assert b_bm.u == a_bm.u
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

            b_bm = a_bm.__floordiv__(mult)
            b_full = a_full.__floordiv__(mult)
            assert b_bm.l == a_bm.l
            assert b_bm.u == a_bm.u
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

            b_bm = a_bm.__div__(mult)
            b_full = a_full.__div__(mult)
            assert b_bm.l == a_bm.l
            assert b_bm.u == a_bm.u
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

            b_bm = a_bm.__truediv__(mult)
            b_full = a_full.__truediv__(mult)
            assert b_bm.l == a_bm.l
            assert b_bm.u == a_bm.u
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

            with self.assertRaises(TypeError):
                a_bm // a_bm
            with self.assertRaises(TypeError):
                a_bm / a_bm
            with self.assertRaises(TypeError):
                1.0 // a_bm
            with self.assertRaises(TypeError):
                1.0 / a_bm

    def test_BandMat_imul(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mult = randn()
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            array_mem = get_array_mem(a_bm.data)

            a_bm *= mult
            a_full *= mult
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data) == array_mem

            with self.assertRaises(TypeError):
                a_bm *= a_bm

    def test_BandMat_various_idivs(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mult = randn()

            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            array_mem = get_array_mem(a_bm.data)
            a_bm //= mult
            a_full //= mult
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data) == array_mem

            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            array_mem = get_array_mem(a_bm.data)
            a_bm /= mult
            a_full /= mult
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data) == array_mem

            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            array_mem = get_array_mem(a_bm.data)
            a_bm.__ifloordiv__(mult)
            a_full.__ifloordiv__(mult)
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data) == array_mem

            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            array_mem = get_array_mem(a_bm.data)
            a_bm.__idiv__(mult)
            a_full.__idiv__(mult)
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data) == array_mem

            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            array_mem = get_array_mem(a_bm.data)
            a_bm.__itruediv__(mult)
            a_full.__itruediv__(mult)
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data) == array_mem

            with self.assertRaises(TypeError):
                a_bm //= a_bm
            with self.assertRaises(TypeError):
                a_bm /= a_bm

    def test_BandMat_reverse_view(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()

            b_bm = a_bm.reverse_view()
            b_full = a_full[::-1, ::-1]
            assert_allclose(b_bm.full(), b_full)
            assert b_bm.data.base is a_bm.data
            if size > 0:
                assert np.may_share_memory(b_bm.data, a_bm.data)

    def test_BandMat_sub_matrix_view(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            start = randint(size + 1)
            end = randint(size + 1)
            if start > end:
                start, end = end, start
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()

            b_bm = a_bm.sub_matrix_view(start, end)
            b_full = a_full[start:end, start:end]
            assert_allclose(b_bm.full(), b_full)
            assert b_bm.data.base is a_bm.data
            if end > start:
                assert np.may_share_memory(b_bm.data, a_bm.data)

    def test_BandMat_embed_as_sub_matrix(self, its = 100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            start = randint(size + 1)
            end = randint(size + 1)
            if start > end:
                start, end = end, start
            a_bm = gen_BandMat(end - start)
            a_full = a_bm.full()

            b_bm = a_bm.embed_as_sub_matrix(start, size)
            b_full = np.zeros((size, size))
            b_full[start:end, start:end] = a_full
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

    def test_zeros(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])

            mat_bm = bm.zeros(l, u, size)
            assert mat_bm.l == l
            assert mat_bm.u == u
            assert_allequal(mat_bm.full(), np.zeros((size, size)))

    def test_from_full(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            mat_full = gen_BandMat(size).full()
            zero_outside_band = np.all(fl.band_ec(l, u, mat_full) == mat_full)

            if zero_outside_band:
                mat_bm = bm.from_full(l, u, mat_full)
                assert mat_bm.l == l
                assert mat_bm.u == u
                assert_allequal(mat_bm.full(), mat_full)
                assert not np.may_share_memory(mat_bm.data, mat_full)
            else:
                self.assertRaises(AssertionError, bm.from_full, l, u, mat_full)

    def test_band_c_bm(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            mat_rect = randn(l + u + 1, size)

            mat_bm = bm.band_c_bm(l, u, mat_rect)

            mat_full_good = fl.band_c(l, u, mat_rect)
            assert_allequal(mat_bm.full(), mat_full_good)
            assert not np.may_share_memory(mat_bm.data, mat_rect)

    def test_band_e_bm(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_bm = gen_BandMat(size)
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])

            mat_rect = bm.band_e_bm(l, u, mat_bm)

            mat_rect_good = fl.band_e(l, u, mat_bm.full())
            assert_allequal(mat_rect, mat_rect_good)
            assert not np.may_share_memory(mat_rect, mat_bm.data)

    def test_band_ec_bm_view(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])

            b_bm = bm.band_ec_bm_view(l, u, a_bm)

            b_full_good = fl.band_ec(l, u, a_bm.full())
            assert_allequal(b_bm.full(), b_full_good)
            assert b_bm.data.base is a_bm.data
            if size > 0:
                assert np.may_share_memory(b_bm.data, a_bm.data)

    def test_band_ec_bm(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])

            b_bm = bm.band_ec_bm(l, u, a_bm)

            b_full_good = fl.band_ec(l, u, a_bm.full())
            assert_allequal(b_bm.full(), b_full_good)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

    def test_band_e_bm_common(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            num_bms = randint(5)
            mat_bms = [ gen_BandMat(size) for _ in range(num_bms) ]
            if num_bms > 0:
                l = max([ mat_bm.l for mat_bm in mat_bms ])
                u = max([ mat_bm.u for mat_bm in mat_bms ])

            mat_rects = bm.band_e_bm_common(*mat_bms)
            for mat_bm, mat_rect in zip(mat_bms, mat_rects):
                assert_allclose(mat_rect, bm.band_e_bm(l, u, mat_bm))

    def test_diag(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])

            vec = randn(size)
            mat_bm = bm.diag(vec)
            assert isinstance(mat_bm, bm.BandMat)
            assert_allequal(mat_bm.full(), np.diag(vec))
            assert mat_bm.data.base is vec
            if size > 0:
                assert np.may_share_memory(mat_bm.data, vec)

            mat_bm = gen_BandMat(size)
            vec = bm.diag(mat_bm)
            assert_allequal(vec, np.diag(mat_bm.full()))
            assert vec.base is mat_bm.data
            if size > 0:
                assert np.may_share_memory(vec, mat_bm.data)

if __name__ == '__main__':
    unittest.main()



"""Tests for operations involving the bands of square matrices."""

# Copyright 2013 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

from bandmat.testhelp import assert_allclose, assert_allequal, get_array_mem

import bandmat.full as fl

import unittest
import numpy as np
import random
from numpy.random import randn, randint

class TestFull(unittest.TestCase):
    def test_band_c_basis(self, its = 100):
        """Checks band_c behaves correctly on canonical basis matrices."""
        for it in range(its):
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            # size >= 1 (there are no canonical basis matrices if size == 0)
            size = random.choice([1, 2, randint(1, 10), randint(1, 100)])

            # pick a random canonical basis matrix
            i = randint(-u, l + 1)
            j = randint(size)
            mat_rect = np.zeros((l + u + 1, size))
            mat_rect[u + i, j] = 1.0

            mat_full = fl.band_c(l, u, mat_rect)

            k = i + j
            mat_full_good = np.zeros((size, size))
            if 0 <= k < size:
                mat_full_good[k, j] = 1.0

            assert_allequal(mat_full, mat_full_good)

    def test_band_c_linear(self, its = 100):
        """Checks band_c is linear."""
        for it in range(its):
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])

            # check additive
            mat_rect1 = randn(l + u + 1, size)
            mat_rect2 = randn(l + u + 1, size)
            assert_allclose(
                fl.band_c(l, u, mat_rect1 + mat_rect2),
                fl.band_c(l, u, mat_rect1) + fl.band_c(l, u, mat_rect2)
            )

            # check homogeneous
            mat_rect = randn(l + u + 1, size)
            mult = random.choice([0.0, randn(), randn(), randn()])
            assert_allclose(
                fl.band_c(l, u, mat_rect * mult),
                fl.band_c(l, u, mat_rect) * mult
            )

            # check output is a newly-created array
            mat_full = fl.band_c(l, u, mat_rect)
            assert not np.may_share_memory(mat_full, mat_rect)

    def test_band_e_basis(self, its = 100):
        """Checks band_e behaves correctly on canonical basis matrices."""
        for it in range(its):
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            # size >= 1 (there are no canonical basis matrices if size == 0)
            size = random.choice([1, 2, randint(1, 10), randint(1, 100)])

            # pick a random canonical basis matrix
            k = randint(size)
            j = randint(size)
            mat_full = np.zeros((size, size))
            mat_full[k, j] = 1.0

            mat_rect = fl.band_e(l, u, mat_full)

            i = k - j
            mat_rect_good = np.zeros((l + u + 1, size))
            if -u <= i <= l:
                mat_rect_good[u + i, j] = 1.0

            assert_allequal(mat_rect, mat_rect_good)

    def test_band_e_linear(self, its = 100):
        """Checks band_e is linear."""
        for it in range(its):
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])

            # check additive
            mat_full1 = randn(size, size)
            mat_full2 = randn(size, size)
            assert_allclose(
                fl.band_e(l, u, mat_full1 + mat_full2),
                fl.band_e(l, u, mat_full1) + fl.band_e(l, u, mat_full2)
            )

            # check homogeneous
            mat_full = randn(size, size)
            mult = random.choice([0.0, randn(), randn(), randn()])
            assert_allclose(
                fl.band_e(l, u, mat_full * mult),
                fl.band_e(l, u, mat_full) * mult
            )

            # check output is a newly-created array
            mat_rect = fl.band_e(l, u, mat_full)
            assert not np.may_share_memory(mat_rect, mat_full)

    def test_zero_extra_entries(self, its = 100):
        """Checks zero_extra_entries against its equivalent definition."""
        for it in range(its):
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_rect = randn(l + u + 1, size)
            mat_rect_good = mat_rect.copy()
            array_mem = get_array_mem(mat_rect)

            fl.zero_extra_entries(l, u, mat_rect)
            mat_rect_good[:] = fl.band_e(l, u, fl.band_c(l, u, mat_rect_good))
            assert_allequal(mat_rect, mat_rect_good)
            assert get_array_mem(mat_rect) == array_mem

    def test_band_ce(self, its = 100):
        """Checks band_ce against its definition and required properties."""
        for it in range(its):
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_rect = randn(l + u + 1, size)

            mat_rect_new = fl.band_ce(l, u, mat_rect)
            mat_rect_new_good = fl.band_e(l, u, fl.band_c(l, u, mat_rect))
            assert_allequal(mat_rect_new, mat_rect_new_good)
            assert not np.may_share_memory(mat_rect_new, mat_rect)

            # check idempotent
            assert_allequal(
                fl.band_ce(l, u, mat_rect_new),
                mat_rect_new
            )

    def test_band_ec(self, its = 100):
        """Checks band_ec against its definition and required properties."""
        for it in range(its):
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_full = randn(size, size)

            mat_full_new = fl.band_ec(l, u, mat_full)
            mat_full_new_good = fl.band_c(l, u, fl.band_e(l, u, mat_full))
            assert_allequal(mat_full_new, mat_full_new_good)
            assert not np.may_share_memory(mat_full_new, mat_full)

            # check idempotent
            assert_allequal(
                fl.band_ec(l, u, mat_full_new),
                mat_full_new
            )

    def test_band_cTe(self, its = 100):
        """Checks band_cTe against its definition and required properties."""
        for it in range(its):
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_rect = randn(l + u + 1, size)

            mat_rect_new = fl.band_cTe(l, u, mat_rect)
            mat_rect_new_good = fl.band_e(u, l, fl.band_c(l, u, mat_rect).T)
            assert_allequal(mat_rect_new, mat_rect_new_good)
            assert not np.may_share_memory(mat_rect_new, mat_rect)

            # check a property to do with doing band_cTe twice
            assert_allequal(
                fl.band_cTe(u, l, mat_rect_new),
                fl.band_ce(l, u, mat_rect)
            )

            # check version that uses pre-specified target
            mat_rect_new2 = np.empty((l + u + 1, size))
            array_mem = get_array_mem(mat_rect, mat_rect_new2)
            ret = fl.band_cTe(l, u, mat_rect, target_rect = mat_rect_new2)
            self.assertIsNone(ret)
            assert_allequal(mat_rect_new2, mat_rect_new)
            assert get_array_mem(mat_rect, mat_rect_new2) == array_mem

if __name__ == '__main__':
    unittest.main()



"""Tests for helper functions for testing."""

# Copyright 2013 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

import bandmat as bm
import bandmat.full as fl
import bandmat.testhelp as th

import unittest
import doctest
import numpy as np
import random
from numpy.random import randn, randint

def randBool():
    return randint(0, 2) == 0

def gen_array(ranks = [0, 1, 2, 3]):
    rank = random.choice(ranks)
    shape = tuple([ randint(5) for _ in range(rank) ])
    return np.asarray(randn(*shape))

def gen_BandMat_simple(size):
    """Generates a random BandMat."""
    l = random.choice([0, 1, randint(0, 10)])
    u = random.choice([0, 1, randint(0, 10)])
    data = randn(l + u + 1, size)
    transposed = randBool()
    return bm.BandMat(l, u, data, transposed = transposed)

def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(th))
    return tests

class TestTestHelp(unittest.TestCase):
    def test_assert_allclose(self):
        a0 = np.array([2.0, 3.0, 4.0])
        a1 = np.array([2.0, 3.0, 4.0])
        a2 = np.array([2.0, 3.0])
        a3 = np.array([2.0, 3.0, 5.0])
        a4 = np.array([[2.0, 3.0, 4.0]])
        th.assert_allclose(a0, a0)
        th.assert_allclose(a0, a1)
        self.assertRaises(AssertionError, th.assert_allclose, a0, a2)
        self.assertRaises(AssertionError, th.assert_allclose, a0, a3)
        self.assertRaises(AssertionError, th.assert_allclose, a0, a4)

    def test_assert_allequal(self):
        a0 = np.array([2.0, 3.0, 4.0])
        a1 = np.array([2.0, 3.0, 4.0])
        a2 = np.array([2.0, 3.0])
        a3 = np.array([2.0, 3.0, 5.0])
        a4 = np.array([[2.0, 3.0, 4.0]])
        th.assert_allequal(a0, a0)
        th.assert_allequal(a0, a1)
        self.assertRaises(AssertionError, th.assert_allequal, a0, a2)
        self.assertRaises(AssertionError, th.assert_allequal, a0, a3)
        self.assertRaises(AssertionError, th.assert_allequal, a0, a4)

    def test_randomize_extra_entries(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])

            mat_rect = randn(l + u + 1, size)
            assert np.all(mat_rect != 0.0)
            fl.zero_extra_entries(l, u, mat_rect)
            th.randomize_extra_entries(l, u, mat_rect)
            assert np.all(mat_rect != 0.0)

            mat_rect = np.zeros((l + u + 1, size))
            assert np.all(mat_rect == 0.0)
            th.randomize_extra_entries(l, u, mat_rect)
            fl.zero_extra_entries(l, u, mat_rect)
            assert np.all(mat_rect == 0.0)

    def test_randomize_extra_entries_bm(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_bm = gen_BandMat_simple(size)

            mat_full = mat_bm.full()
            th.randomize_extra_entries_bm(mat_bm)
            th.assert_allequal(mat_bm.full(), mat_full)

    def test_get_array_mem(self, its = 50):
        # (FIXME : these are not great tests, since not familiar enough with
        #   numpy internals to know what sorts of changes in memory layout are
        #   possible and how they might arise in a realistic program)
        for it in range(its):
            x = gen_array()
            array_mem = th.get_array_mem(x)
            x *= 2.0
            assert th.get_array_mem(x) == array_mem

            x = gen_array()
            array_mem = th.get_array_mem(x)
            x.shape = x.shape + (1,)
            assert th.get_array_mem(x) != array_mem

            x = gen_array(ranks = [1, 2, 3])
            array_mem = th.get_array_mem(x)
            shape = x.shape
            strides = x.strides
            shape_new = x.T.shape
            strides_new = x.T.strides
            if np.prod(shape_new) != 0:
                x.shape = shape_new
                x.strides = strides_new
                if shape_new != shape or strides_new != strides:
                    assert th.get_array_mem(x) != array_mem

if __name__ == '__main__':
    unittest.main()



"""Assorted helpful functions."""

# Copyright 2013 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

cimport numpy as cnp
cimport cython

cnp.import_array()
cnp.import_ufunc()

def fancy_plus_equals(cnp.ndarray[cnp.int64_t, ndim=1] targetIndexSeq,
                      cnp.ndarray[cnp.float64_t, ndim=1] source,
                      cnp.ndarray[cnp.float64_t, ndim=1] target):
    """Implements a += method with fancy indexing.

    Does what you might expect
        target[targetIndexSeq] += source
    to do.
    """
    cdef unsigned long sourceSize

    sourceSize = source.shape[0]
    assert targetIndexSeq.shape[0] == sourceSize

    cdef unsigned long sourceIndex
    cdef long targetIndex

    for sourceIndex in range(sourceSize):
        targetIndex = targetIndexSeq[sourceIndex]
        target[targetIndex] += source[sourceIndex]

    return

def fancy_plus_equals_2d(cnp.ndarray[cnp.int64_t, ndim=1] targetIndexSeq,
                         cnp.ndarray[cnp.float64_t, ndim=2] source,
                         cnp.ndarray[cnp.float64_t, ndim=2] target):
    """Implements a += method with fancy indexing.

    Does what you might expect
        target[targetIndexSeq] += source
    to do.
    """
    cdef unsigned long sourceSize
    cdef unsigned long size1

    sourceSize = source.shape[0]
    assert targetIndexSeq.shape[0] == sourceSize
    size1 = source.shape[1]
    assert target.shape[1] == size1

    cdef unsigned long sourceIndex
    cdef long targetIndex
    cdef unsigned long index1

    for sourceIndex in range(sourceSize):
        targetIndex = targetIndexSeq[sourceIndex]
        for index1 in range(size1):
            target[targetIndex, index1] += source[sourceIndex, index1]

    return



"""Tests for linear algebra operations for banded matrices."""

# Copyright 2013 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

from bandmat.testhelp import assert_allclose, randomize_extra_entries_bm
from bandmat.test_core import gen_BandMat

import bandmat as bm
import bandmat.full as fl
import bandmat.linalg as bla

import unittest
import numpy as np
import numpy.linalg as la
import scipy.linalg as sla
import random
from numpy.random import randn, randint

def randBool():
    return randint(0, 2) == 0

def gen_symmetric_BandMat(size, depth = None):
    if depth is None:
        depth = random.choice([0, 1, randint(0, 10)])
    a_bm = gen_BandMat(size, l = depth, u = depth)
    b_bm = a_bm + a_bm.T
    randomize_extra_entries_bm(b_bm)
    return b_bm

def gen_pos_def_BandMat(size, depth = None, contribRank = 2):
    """Generates a random positive definite BandMat."""
    assert contribRank >= 0
    if depth is None:
        depth = random.choice([0, 1, randint(0, 10)])
    mat_bm = bm.zeros(depth, depth, size)
    for _ in range(contribRank):
        diff = randint(0, depth + 1)
        chol_bm = gen_BandMat(size, l = depth - diff, u = diff)
        bm.dot_mm_plus_equals(chol_bm, chol_bm.T, mat_bm)
    transposed = randBool()
    if transposed:
        mat_bm = mat_bm.T
    randomize_extra_entries_bm(mat_bm)
    return mat_bm

def gen_chol_factor_BandMat(size, depth = None, contribRank = 2):
    """Generates a random Cholesky factor BandMat.

    This works by generating a random positive definite matrix and then
    computing its Cholesky factor, since using a random matrix as a Cholesky
    factor seems to often lead to ill-conditioned matrices.
    """
    mat_bm = gen_pos_def_BandMat(size, depth = depth,
                                 contribRank = contribRank)
    chol_bm = bla.cholesky(mat_bm, lower = randBool())
    if randBool():
        chol_bm = chol_bm.T
    assert chol_bm.l == 0 or chol_bm.u == 0
    assert chol_bm.l + chol_bm.u == mat_bm.l
    randomize_extra_entries_bm(chol_bm)
    return chol_bm

class TestLinAlg(unittest.TestCase):
    def test_cholesky(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_bm = gen_pos_def_BandMat(size)
            depth = mat_bm.l
            lower = randBool()
            alternative = randBool()

            chol_bm = bla.cholesky(mat_bm, lower = lower,
                                   alternative = alternative)
            assert chol_bm.l == (depth if lower else 0)
            assert chol_bm.u == (0 if lower else depth)
            assert not np.may_share_memory(chol_bm.data, mat_bm.data)

            if lower != alternative:
                mat_bm_again = bm.dot_mm(chol_bm, chol_bm.T)
            else:
                mat_bm_again = bm.dot_mm(chol_bm.T, chol_bm)
            assert_allclose(mat_bm_again.full(), mat_bm.full())

    def test_cho_solve(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            b = randn(size)
            chol_bm = gen_chol_factor_BandMat(size)
            depth = chol_bm.l + chol_bm.u
            lower = (chol_bm.u == 0)
            chol_full = chol_bm.full()

            x = bla.cho_solve(chol_bm, b)
            if size == 0:
                x_good = np.zeros((size,))
            else:
                x_good = sla.cho_solve((chol_full, lower), b)
            assert_allclose(x, x_good)
            assert not np.may_share_memory(x, chol_bm.data)
            assert not np.may_share_memory(x, b)

    def test_solve(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            b = randn(size)
            # the below tries to ensure the matrix is well-conditioned
            a_bm = gen_BandMat(size) + bm.diag(np.ones((size,)) * 10.0)
            a_full = a_bm.full()

            x = bla.solve(a_bm, b)
            if size == 0:
                x_good = np.zeros((size,))
            else:
                x_good = sla.solve(a_full, b)
            assert_allclose(x, x_good)
            assert not np.may_share_memory(x, a_bm.data)
            assert not np.may_share_memory(x, b)

    def test_solveh(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            b = randn(size)
            a_bm = gen_pos_def_BandMat(size)
            a_full = a_bm.full()

            x = bla.solveh(a_bm, b)
            if size == 0:
                x_good = np.zeros((size,))
            else:
                x_good = sla.solve(a_full, b, sym_pos = True)
            assert_allclose(x, x_good)
            assert not np.may_share_memory(x, a_bm.data)
            assert not np.may_share_memory(x, b)

    def test_band_of_inverse_from_chol(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            chol_bm = gen_chol_factor_BandMat(size)
            depth = chol_bm.l + chol_bm.u

            band_of_inv_bm = bla.band_of_inverse_from_chol(chol_bm)
            assert not np.may_share_memory(band_of_inv_bm.data, chol_bm.data)

            mat_bm = (bm.dot_mm(chol_bm, chol_bm.T) if chol_bm.u == 0
                      else bm.dot_mm(chol_bm.T, chol_bm))
            band_of_inv_full_good = fl.band_ec(
                depth, depth,
                np.eye(0, 0) if size == 0 else la.inv(mat_bm.full())
            )
            assert_allclose(band_of_inv_bm.full(), band_of_inv_full_good)

    def test_band_of_inverse(self, its = 50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_bm = gen_pos_def_BandMat(size)
            depth = mat_bm.l

            band_of_inv_bm = bla.band_of_inverse(mat_bm)
            assert not np.may_share_memory(band_of_inv_bm.data, mat_bm.data)

            band_of_inv_full_good = fl.band_ec(
                depth, depth,
                np.eye(0, 0) if size == 0 else la.inv(mat_bm.full())
            )
            assert_allclose(band_of_inv_bm.full(), band_of_inv_full_good)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
"""A simple example of how the bandmat package may be used."""

import numpy as np
import bandmat as bm

a_bm = bm.BandMat(
    1, 1,
    np.array([
        [0.0, 0.2, 0.3, 0.4, 0.5],
        [1.0, 0.9, 1.1, 0.8, 1.3],
        [0.3, 0.1, 0.5, 0.6, 0.0],
    ])
)

b_bm = bm.BandMat(
    1, 0,
    np.array([
        [1.0, 0.9, 1.1, 0.8, 1.3],
        [0.3, 0.1, 0.5, 0.6, 0.0],
    ])
)

c_bm = bm.dot_mm(a_bm.T, b_bm)

d_bm = a_bm + b_bm

a_full = a_bm.full()
b_full = b_bm.full()
c_full = c_bm.full()
d_full = d_bm.full()

print 'a_full:'
print a_full
print
print 'b_full:'
print b_full
print
print 'np.dot(a_full.T, b_full):'
print c_full
print
print 'a_full + b_full:'
print d_full
print

c_full_again = np.dot(a_full.T, b_full)
assert np.allclose(c_full_again, c_full)

d_full_again = a_full + b_full
assert np.allclose(d_full_again, d_full)



#!/usr/bin/python
"""A distutils-based script for distributing and installing bandmat."""

# Copyright 2013 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

import os
from distutils.core import setup
from distutils.extension import Extension
from distutils.command.sdist import sdist as _sdist

cython_locs = [
    ('bandmat', 'full'),
    ('bandmat', 'core'),
    ('bandmat', 'tensor'),
    ('bandmat', 'linalg'),
    ('bandmat', 'misc'),
]

with open('README.rst') as readmeFile:
    long_description = readmeFile.read()

# see "A note on setup.py" in README.rst for an explanation of the dev file
dev_mode = os.path.exists('dev')

if dev_mode:
    from Cython.Distutils import build_ext
    from Cython.Build import cythonize

    class sdist(_sdist):
        """A cythonizing sdist command.

        This class is a custom sdist command which ensures all cython-generated
        C files are up-to-date before running the conventional sdist command.
        """
        def run(self):
            cythonize([ os.path.join(*loc)+'.pyx' for loc in cython_locs ])
            _sdist.run(self)

    cmdclass = {'build_ext': build_ext, 'sdist': sdist}
    ext_modules = [ Extension('.'.join(loc), [os.path.join(*loc)+'.pyx'])
                    for loc in cython_locs ]
else:
    cmdclass = {}
    ext_modules = [ Extension('.'.join(loc), [os.path.join(*loc)+'.c'])
                    for loc in cython_locs ]

setup(
    name = 'bandmat',
    version = '0.2',
    description = 'A banded matrix library for python.',
    url = 'http://github.com/MattShannon/bandmat',
    author = 'Matt Shannon',
    author_email = 'matt.shannon@cantab.net',
    license = '3-clause BSD (see License file)',
    packages = ['bandmat'],
    long_description = long_description,
    cmdclass = cmdclass,
    ext_modules = ext_modules,
)



"""A banded matrix library for python.

This package provides a simple banded matrix library for python.
It supports banded matrix-vector and matrix-matrix multiplication, converting
between full and banded matrix representations, and certain linear algebra
operations on banded matrices.

Overview
--------

A banded matrix is a matrix where only the diagonal, a number of superdiagonals
and a number of subdiagonals are non-zero.
The well-known BLAS interface and LAPACK library for linear algebra define
several banded matrix operations, and some of these, such as banded Cholesky
decomposition, are wrapped in the excellent python package scipy, specifically
in scipy.linalg.
The bandmat package re-uses the banded matrix representation used by BLAS,
LAPACK and scipy.linalg, wrapping it in a lightweight class for ease of use.
See the docstring for the BandMat class for full details of the representation
used.

The bandmat package provides:

- a lightweight class wrapping the LAPACK-style banded matrix representation.
  This class keeps track of things like bandwidths to allow a more direct
  coding style when working with banded matrices.
- some basic banded matrix operations not present in scipy.
  For example, banded matrix-vector multiplication is defined by BLAS but not
  wrapped by scipy, and banded matrix-matrix multiplication is not defined in
  BLAS or in scipy.
  The bandmat package contains C implementations of these operations written in
  cython.
- helper functions for converting between full and banded matrix
  representations.
- certain linear algebra operations on banded matrices.
- an implementation of a fancy indexed += function for numpy arrays.
  This is included for (the author's) convenience and is not directly related
  to banded matrix manipulation.

Only square banded matrices are supported by this package.

Examples
--------

Define a banded matrix:
>>> import numpy as np
>>> import bandmat as bm
>>> a_bm = bm.BandMat(1, 1, np.arange(15, dtype=np.float64).reshape((3, 5)))
>>> print a_bm.data
[[  0.   1.   2.   3.   4.]
 [  5.   6.   7.   8.   9.]
 [ 10.  11.  12.  13.  14.]]

Get the equivalent numpy array:
>>> print a_bm.full()
[[  5.   1.   0.   0.   0.]
 [ 10.   6.   2.   0.   0.]
 [  0.  11.   7.   3.   0.]
 [  0.   0.  12.   8.   4.]
 [  0.   0.   0.  13.   9.]]

Take the transpose:
>>> print a_bm.T.full()
[[  5.  10.   0.   0.   0.]
 [  1.   6.  11.   0.   0.]
 [  0.   2.   7.  12.   0.]
 [  0.   0.   3.   8.  13.]
 [  0.   0.   0.   4.   9.]]

Banded matrix addition:
>>> b_bm = bm.BandMat(0, 1, np.arange(10, dtype=np.float64).reshape((2, 5)))
>>> print (a_bm + b_bm).full()
[[ 10.   2.   0.   0.   0.]
 [ 10.  12.   4.   0.   0.]
 [  0.  11.  14.   6.   0.]
 [  0.   0.  12.  16.   8.]
 [  0.   0.   0.  13.  18.]]

Banded matrix multiplication:
>>> b_bm = bm.BandMat(0, 1, np.arange(10, dtype=np.float64).reshape((2, 5)))
>>> print bm.dot_mm(a_bm, b_bm).full()
[[  25.   11.    2.    0.    0.]
 [  50.   46.   26.    6.    0.]
 [   0.   66.   71.   45.   12.]
 [   0.    0.   84.  100.   68.]
 [   0.    0.    0.  104.  133.]]
"""

# Copyright 2013 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

from bandmat.core import *
from bandmat.tensor import *

__version__ = '0.2'



"""Operations involving the bands of square matrices.

This module provides operations involving the bands of square matrices which
are stored as numpy arrays.
For example it provides functions to construct a square matrix with given band,
and to extract the band from an arbitrary square matrix.
Though this is closely related to the representation of square banded matrices
used by the `BandMat` class in the `core` module, it is logically distinct.
"""

# Copyright 2013 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

import numpy as np

cimport numpy as cnp
cimport cython

cnp.import_array()
cnp.import_ufunc()

@cython.boundscheck(False)
def band_c(long l, long u, cnp.ndarray[cnp.float64_t, ndim=2] mat_rect):
    """Constructs a square banded matrix from its band.

    Given a rectangular numpy array `mat_rect`, this function returns a square
    numpy array `mat_full` with its `u` superdiagonals, diagonal and `l`
    subdiagonals given given by the rows of `mat_rect`.
    The part of each column of `mat_full` that lies within the band contains
    the same entries as the corresponding column of `mat_rect`.

    Not all the entries of `mat_rect` affect `mat_full`.
    The (l, u)-extra entries of a rectangular matrix `mat_rect` are defined as
    the entries which have no effect on the result of `band_c(l, u, mat_rect)`.
    They lie in the upper-left and bottom-right corners of `mat_rect`.
    """
    assert l >= 0
    assert u >= 0
    assert mat_rect.shape[0] == l + u + 1

    cdef long size
    cdef cnp.ndarray[cnp.float64_t, ndim=2] mat_full

    size = mat_rect.shape[1]
    mat_full = np.zeros((size, size))

    cdef long i
    cdef unsigned long row
    cdef unsigned long j

    for i in range(-u, l + 1):
        row = u + i
        for j in range(max(0, -i), max(0, size + min(0, -i))):
            mat_full[j + i, j] = mat_rect[row, j]

    return mat_full

@cython.boundscheck(False)
def band_e(long l, long u, cnp.ndarray[cnp.float64_t, ndim=2] mat_full):
    """Extracts a band of a square matrix.

    Given a square numpy array `mat_full`, returns a rectangular numpy array
    `mat_rect` with rows corresponding to the `u` superdiagonals, the diagonal
    and the `l` subdiagonals of `mat_full`.
    The square matrix is "collapsed column-wise", i.e. each column of
    `mat_rect` contains the same entries as the part of the corresponding
    column of `mat_full` that lies within the band defined by (l, u).
    The extra entries in the corners of `mat_rect` which do not correspond to
    any entry in `mat_full` are set to zero.
    """
    assert l >= 0
    assert u >= 0
    assert mat_full.shape[1] == mat_full.shape[0]

    cdef long size
    cdef cnp.ndarray[cnp.float64_t, ndim=2] mat_rect

    size = mat_full.shape[0]
    mat_rect = np.empty((l + u + 1, size))

    cdef long i
    cdef unsigned long row
    cdef unsigned long j

    for i in range(-u, l + 1):
        row = u + i
        for j in range(size):
            # "j + i < size" below uses wraparound on unsigned long to actually
            #   check that "0 <= j + i < size"
            mat_rect[row, j] = mat_full[j + i, j] if j + i < size else 0.0

    return mat_rect

@cython.boundscheck(False)
def zero_extra_entries(long l, long u,
                       cnp.ndarray[cnp.float64_t, ndim=2] mat_rect):
    """Zeroes the extra entries of a rectangular matrix.

    See the docstring for `band_c` for the definition of extra entries.

    The statement `zero_extra_entries(l, u, mat_rect)` is equivalent to:

        mat_rect[:] = band_e(l, u, band_c(l, u, mat_rect))

    N.B. in-place, i.e. mutates `mat_rect`.
    """
    assert l >= 0
    assert u >= 0
    assert mat_rect.shape[0] == l + u + 1

    cdef long size

    size = mat_rect.shape[1]

    cdef long i
    cdef unsigned long row
    cdef unsigned long j

    for i in range(-u, 0):
        row = u + i
        for j in range(0, min(size, -i)):
            mat_rect[row, j] = 0.0
    for i in range(1, l + 1):
        row = u + i
        for j in range(max(0, size - i), size):
            mat_rect[row, j] = 0.0

    return

def band_ce(l, u, mat_rect):
    """Effectively applies `band_c` then `band_e`.

    The combined operation has the effect of zeroing the extra entries in (a
    copy of) the rectangular matrix `mat_rect`.

    The expression `band_ce(l, u, mat_rect)` is equivalent to:

        band_e(l, u, band_c(l, u, mat_rect))
    """
    mat_rect_new = mat_rect.copy()
    zero_extra_entries(l, u, mat_rect_new)
    return mat_rect_new

def band_ec(l, u, mat_full):
    """Effectively applies `band_e` then `band_c`.

    The combined operation has the effect of zeroing the entries outside the
    band specified by `l` and `u` in (a copy of) the square matrix `mat_full`.

    The expression `band_ec(l, u, mat_full)` is equivalent to:

        band_c(l, u, band_e(l, u, mat_full))
    """
    return band_c(l, u, band_e(l, u, mat_full))

@cython.boundscheck(False)
def band_cTe(long l, long u,
             cnp.ndarray[cnp.float64_t, ndim=2] mat_rect,
             cnp.ndarray[cnp.float64_t, ndim=2] target_rect = None):
    """Effectively applies `band_c` then transpose then `band_e`.

    The expression `band_cTe(l, u, mat_rect)` is equivalent to:

        band_e(u, l, band_c(l, u, mat_rect).T)
    """
    assert l >= 0
    assert u >= 0
    assert mat_rect.shape[0] == l + u + 1

    cdef long size
    cdef long target_given = (target_rect is not None)

    size = mat_rect.shape[1]
    if not target_given:
        target_rect = np.empty((l + u + 1, size))

    cdef long i
    cdef unsigned long row, row_new
    cdef unsigned long j

    for i in range(-u, l + 1):
        row = u + i
        row_new = l - i
        for j in range(size):
            # "j - i < size" below uses wraparound on unsigned long to actually
            #   check that "0 <= j - i < size"
            target_rect[row_new, j] = (mat_rect[row, j - i] if j - i < size
                                       else 0.0)

    if target_given:
        return
    else:
        return target_rect



"""Tests for assorted helpful functions."""

# Copyright 2013 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

from bandmat.testhelp import assert_allclose, assert_allequal, get_array_mem

from bandmat.misc import fancy_plus_equals, fancy_plus_equals_2d

import unittest
import numpy as np
import random
from numpy.random import randn, randint

def randBool():
    return randint(0, 2) == 0

class TestMisc(unittest.TestCase):
    def test_fancy_plus_equals(self, its = 100):
        for it in range(its):
            sourceSize = random.choice([0, 1, randint(10), randint(100)])
            targetSize = random.choice([1, randint(1, 10), randint(1, 100)])
            source = randn(sourceSize)
            target = randn(targetSize)
            targetIndexSeq = np.array([ randint(targetSize)
                                        for _ in range(sourceSize) ],
                                      dtype = np.int64)
            array_mem = get_array_mem(targetIndexSeq, source, target)

            targetGood = target.copy()
            for sourceIndex, targetIndex in enumerate(targetIndexSeq):
                targetGood[targetIndex] += source[sourceIndex]

            fancy_plus_equals(targetIndexSeq, source, target)

            assert_allclose(target, targetGood)
            assert get_array_mem(targetIndexSeq, source, target) == array_mem

    def test_fancy_plus_equals_fails(self, its = 100):
        for it in range(its):
            sourceSize = random.choice([1, randint(1, 10), randint(1, 100)])
            targetSize = random.choice([1, randint(1, 10), randint(1, 100)])
            source = randn(sourceSize)
            target = randn(targetSize)
            targetIndexSeq = np.array([ randint(targetSize)
                                        for _ in range(sourceSize) ],
                                      dtype = np.int64)
            targetIndexSeq[randint(sourceSize)] = (
                (-targetSize - 1 - randint(10)) if randBool()
                else targetSize + randint(10)
            )
            self.assertRaises(IndexError,
                              fancy_plus_equals,
                              targetIndexSeq, source, target)

    def test_fancy_plus_equals_2d(self, its = 100):
        for it in range(its):
            sourceSize = random.choice([0, 1, randint(10), randint(100)])
            targetSize = random.choice([1, randint(1, 10), randint(1, 100)])
            size1 = random.choice([0, 1, randint(10)])
            source = randn(sourceSize, size1)
            target = randn(targetSize, size1)
            targetIndexSeq = np.array([ randint(targetSize)
                                        for _ in range(sourceSize) ],
                                      dtype = np.int64)
            array_mem = get_array_mem(targetIndexSeq, source, target)

            targetGood = target.copy()
            for sourceIndex, targetIndex in enumerate(targetIndexSeq):
                targetGood[targetIndex] += source[sourceIndex]

            fancy_plus_equals_2d(targetIndexSeq, source, target)

            assert_allclose(target, targetGood)
            assert get_array_mem(targetIndexSeq, source, target) == array_mem

    def test_fancy_plus_equals_2d_fails(self, its = 100):
        for it in range(its):
            sourceSize = random.choice([1, randint(1, 10), randint(1, 100)])
            targetSize = random.choice([1, randint(1, 10), randint(1, 100)])
            size1 = random.choice([1, randint(1, 10)])
            source = randn(sourceSize, size1)
            target = randn(targetSize, size1)
            targetIndexSeq = np.array([ randint(targetSize)
                                        for _ in range(sourceSize) ],
                                      dtype = np.int64)
            targetIndexSeq[randint(sourceSize)] = (
                (-targetSize - 1 - randint(10)) if randBool()
                else targetSize + randint(10)
            )
            self.assertRaises(IndexError,
                              fancy_plus_equals_2d,
                              targetIndexSeq, source, target)

if __name__ == '__main__':
    unittest.main()



"""Helper functions for testing."""

import numpy as np
from numpy.random import randn

def assert_allclose(actual, desired, rtol = 1e-7, atol = 1e-14,
                    msg = 'items not almost equal'):
    if np.shape(actual) != np.shape(desired):
        raise AssertionError('%s (wrong shape)\n ACTUAL:  %r\n DESIRED: %r' %
                             (msg, actual, desired))
    if not np.allclose(actual, desired, rtol, atol):
        absErr = np.abs(actual - desired)
        relErr = np.abs((actual - desired) / desired)
        raise AssertionError('%s\n ACTUAL:  %r\n DESIRED: %r\n'
                             ' ABS ERR: %r (max %s)\n REL ERR: %r (max %s)' %
                             (msg, actual, desired,
                              absErr, np.max(absErr), relErr, np.max(relErr)))

def assert_allequal(actual, desired, msg = 'items not equal'):
    if np.shape(actual) != np.shape(desired):
        raise AssertionError('%s (wrong shape)\n ACTUAL:  %r\n DESIRED: %r' %
                             (msg, actual, desired))
    if not np.all(actual == desired):
        raise AssertionError('%s\n ACTUAL:  %r\n DESIRED: %r' %
                             (msg, actual, desired))

def randomize_extra_entries(l, u, mat_rect):
    """Randomizes the extra entries of a rectangular matrix.

    See the docstring for `band_c` for the definition of extra entries.

    N.B. in-place, i.e. mutates `mat_rect`.
    """
    assert l >= 0
    assert u >= 0
    assert mat_rect.shape[0] == l + u + 1

    size = mat_rect.shape[1]

    for offset in range(1, u + 1):
        mat_rect[u - offset, 0:min(offset, size)] = randn()
    for offset in range(1, l + 1):
        mat_rect[u + offset, max(size - offset, 0):size] = randn()

def randomize_extra_entries_bm(mat_bm):
    if mat_bm.transposed:
        randomize_extra_entries(mat_bm.u, mat_bm.l, mat_bm.data)
    else:
        randomize_extra_entries(mat_bm.l, mat_bm.u, mat_bm.data)

def get_array_mem(*arrays):
    """Returns a representation of the memory layout of an array.

    This is intended to be used to check whether the memory used by a given
    numpy array, or how this memory is mapped into the logical indices of the
    tensor it represents, changes between two points in time.

    Example usage:
    >>> import numpy as np
    >>> x = np.array([2.0, 3.0, 4.0])
    >>> array_mem = get_array_mem(x)
    >>> # some potentially complicated operation
    >>> x *= 2.0
    >>> assert get_array_mem(x) == array_mem
    """
    return [ array.__array_interface__ for array in arrays ]
